@@ -6,6 +6,7 @@
 #include <sys/time.h>
 
 #include <amqp.h>
+#include <amqp_tcp_socket.h>
 
 #include "connection.h"
 #include "distmeta.h"
@@ -68,7 +69,8 @@ basic_properties_to_PyDict(amqp_basic_properties_t*, PyObject*);
 _PYRMQ_INLINE int
 PyDict_to_basic_properties(PyObject *,
                            amqp_basic_properties_t *,
-                           amqp_connection_state_t );
+                           amqp_connection_state_t,
+                           amqp_pool_t *);
 
 _PYRMQ_INLINE void
 amqp_basic_deliver_to_PyDict(PyObject *, uint64_t, amqp_bytes_t,
@@ -96,8 +98,8 @@ int PyRabbitMQ_HandleAMQError(PyRabbitMQ_Connection *, unsigned int,
 void PyRabbitMQ_SetErr_UnexpectedHeader(amqp_frame_t*);
 int PyRabbitMQ_Not_Connected(PyRabbitMQ_Connection *);
 
-static amqp_table_t PyDict_ToAMQTable(amqp_connection_state_t, PyObject *);
-static amqp_array_t PyList_ToAMQArray(amqp_connection_state_t, PyObject *);
+static amqp_table_t PyDict_ToAMQTable(amqp_connection_state_t, PyObject *, amqp_pool_t *);
+static amqp_array_t PyList_ToAMQArray(amqp_connection_state_t, PyObject *, amqp_pool_t *);
 
 static PyObject* AMQTable_toPyDict(amqp_table_t *table);
 static PyObject* AMQArray_toPyList(amqp_array_t *array);
@@ -210,7 +212,7 @@ AMQArray_SetIntValue(amqp_array_t *array, int value)
 }
 
 static amqp_table_t
-PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src)
+PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src, amqp_pool_t *pool)
 {
     PyObject *dkey = NULL;
     PyObject *dvalue = NULL;
@@ -225,21 +227,20 @@ PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src)
 
     /* allocate new table */
     dst.num_entries = 0;
-    dst.entries = amqp_pool_alloc(&conn->frame_pool,
-                           size * sizeof(amqp_table_entry_t));
+    dst.entries = amqp_pool_alloc(pool, size * sizeof(amqp_table_entry_t));
     while (PyDict_Next(src, &pos, &dkey, &dvalue)) {
 
         if (PyDict_Check(dvalue)) {
             /* Dict */
             AMQTable_SetTableValue(&dst,
                     PyString_AS_AMQBYTES(dkey),
-                    PyDict_ToAMQTable(conn, dvalue));
+                    PyDict_ToAMQTable(conn, dvalue, pool));
         }
         else if (PyList_Check(dvalue)) {
             /* List */
             AMQTable_SetArrayValue(&dst,
                     PyString_AS_AMQBYTES(dkey),
-                    PyList_ToAMQArray(conn, dvalue));
+                    PyList_ToAMQArray(conn, dvalue, pool));
         }
         else if (PyLong_Check(dvalue) || PyInt_Check(dvalue)) {
             /* Int | Long */
@@ -292,7 +293,7 @@ error:
 }
 
 static amqp_array_t
-PyList_ToAMQArray(amqp_connection_state_t conn, PyObject *src)
+PyList_ToAMQArray(amqp_connection_state_t conn, PyObject *src, amqp_pool_t *pool)
 {
     PyObject *dvalue = NULL;
     PY_SIZE_TYPE size = 0;
@@ -305,8 +306,7 @@ PyList_ToAMQArray(amqp_connection_state_t conn, PyObject *src)
 
     /* allocate new array */
     dst.num_entries = 0;
-    dst.entries = amqp_pool_alloc(&conn->frame_pool,
-                           size * sizeof(amqp_field_value_t));
+    dst.entries = amqp_pool_alloc(pool, size * sizeof(amqp_field_value_t));
     for (pos = 0; pos < size; ++pos) {
 
         dvalue = PyList_GetItem(src, pos);
@@ -314,12 +314,12 @@ PyList_ToAMQArray(amqp_connection_state_t conn, PyObject *src)
         if (PyDict_Check(dvalue)) {
             /* Dict */
             AMQArray_SetTableValue(&dst,
-                    PyDict_ToAMQTable(conn, dvalue));
+                    PyDict_ToAMQTable(conn, dvalue, pool));
         }
         else if (PyList_Check(dvalue)) {
             /* List */
             AMQArray_SetArrayValue(&dst,
-                    PyList_ToAMQArray(conn, dvalue));
+                    PyList_ToAMQArray(conn, dvalue, pool));
         }
         else if (PyLong_Check(dvalue) || PyInt_Check(dvalue)) {
             /* Int | Long */
@@ -462,8 +462,8 @@ basic_properties_to_PyDict(amqp_basic_properties_t *props, PyObject *p)
         PyDICT_SETSTR_DECREF(p, "timestamp", value);
     }
     if (props->_flags & AMQP_BASIC_HEADERS_FLAG) {
-        PyDICT_SETSTR_DECREF(p, "headers",
-            AMQTable_toPyDict(&(props->headers)));
+        value = AMQTable_toPyDict(&(props->headers));
+        PyDICT_SETSTR_DECREF(p, "headers", value);
     }
 }
 
@@ -617,7 +617,8 @@ AMQArray_toPyList(amqp_array_t *array)
 _PYRMQ_INLINE int
 PyDict_to_basic_properties(PyObject *p,
                            amqp_basic_properties_t *props,
-                           amqp_connection_state_t conn)
+                           amqp_connection_state_t conn,
+                           amqp_pool_t *pool)
 {
     PyObject *value = NULL;
     props->headers = AMQP_EMPTY_TABLE;
@@ -682,7 +683,7 @@ PyDict_to_basic_properties(PyObject *p,
     }
 
     if ((value = PyDict_GetItemString(p, "headers")) != NULL) {
-        props->headers = PyDict_ToAMQTable(conn, value);
+        props->headers = PyDict_ToAMQTable(conn, value, pool);
         if (PyErr_Occurred())
             return -1;
     }
@@ -761,7 +762,7 @@ int PyRabbitMQ_HandleAMQError(PyRabbitMQ_Connection *self, unsigned int channel,
             snprintf(errorstr, sizeof(errorstr), "%s: %s",
                     context,
                     reply.library_error
-                        ? amqp_error_string(reply.library_error)
+                        ? amqp_error_string2(reply.library_error)
                         : "(end-of-stream)");
             goto connerror;
 
@@ -848,6 +849,7 @@ PyRabbitMQ_ConnectionType_new(PyTypeObject *type,
         self->port = 5672;
         self->sockfd = 0;
         self->connected = 0;
+        self->server_properties = NULL;
         self->callbacks = NULL;
     }
     return self;
@@ -863,6 +865,7 @@ PyRabbitMQ_ConnectionType_dealloc(PyRabbitMQ_Connection *self)
     if (self->weakreflist != NULL)
         PyObject_ClearWeakRefs((PyObject*)self);
     Py_XDECREF(self->callbacks);
+    Py_XDECREF(self->server_properties);
     self->ob_type->tp_free(self);
 }
 
@@ -910,8 +913,8 @@ PyRabbitMQ_ConnectionType_init(PyRabbitMQ_Connection *self,
     self->heartbeat = heartbeat;
     self->weakreflist = NULL;
     self->callbacks = PyDict_New();
-    if (self->callbacks == NULL)
-        return -1;
+    if (self->callbacks == NULL) return -1;
+    self->server_properties = NULL;
 
     return 0;
 }
@@ -923,7 +926,13 @@ PyRabbitMQ_ConnectionType_init(PyRabbitMQ_Connection *self,
 static PyObject*
 PyRabbitMQ_Connection_fileno(PyRabbitMQ_Connection *self)
 {
-    return PyInt_FromLong((long)self->sockfd);
+    if (self->sockfd > 0) {
+        return PyInt_FromLong((long)self->sockfd);
+    }
+    else {
+        PyErr_SetString(PyExc_ValueError, "Socket not connected");
+        return 0;
+    }
 }
 
 
@@ -933,6 +942,8 @@ PyRabbitMQ_Connection_fileno(PyRabbitMQ_Connection *self)
 static PyObject*
 PyRabbitMQ_Connection_connect(PyRabbitMQ_Connection *self)
 {
+    int status;
+    amqp_socket_t *socket = NULL;
     amqp_rpc_reply_t reply;
 
     if (self->connected) {
@@ -941,13 +952,19 @@ PyRabbitMQ_Connection_connect(PyRabbitMQ_Connection *self)
     }
     Py_BEGIN_ALLOW_THREADS;
     self->conn = amqp_new_connection();
-    self->sockfd = amqp_open_socket(self->hostname, self->port);
+    socket = amqp_tcp_socket_new(self->conn);
     Py_END_ALLOW_THREADS;
+
+    if (!socket) goto error;
+    Py_BEGIN_ALLOW_THREADS;
+    status = amqp_socket_open(socket, self->hostname, self->port);
+    Py_END_ALLOW_THREADS;
+    if (status) goto error;
     if (!PyRabbitMQ_HandleError(self->sockfd, "Error opening socket"))
         goto error;
 
     Py_BEGIN_ALLOW_THREADS;
-    amqp_set_sockfd(self->conn, self->sockfd);
+    self->sockfd = amqp_socket_get_sockfd(socket);
     reply = amqp_login(self->conn, self->virtual_host, self->channel_max,
                        self->frame_max, self->heartbeat,
                        AMQP_SASL_METHOD_PLAIN, self->userid, self->password);
@@ -960,6 +977,7 @@ PyRabbitMQ_Connection_connect(PyRabbitMQ_Connection *self)
     self->channel_max = self->conn->channel_max;
     self->frame_max = self->conn->frame_max;
     self->heartbeat = self->conn->heartbeat;
+    self->server_properties = AMQTable_toPyDict(amqp_get_server_properties(self->conn));
     Py_RETURN_NONE;
 error:
     PyRabbitMQ_Connection_close(self);
@@ -982,7 +1000,7 @@ PyRabbitMQ_Connection_close(PyRabbitMQ_Connection *self)
         Py_BEGIN_ALLOW_THREADS
         reply = amqp_connection_close(self->conn, AMQP_REPLY_SUCCESS);
         amqp_destroy_connection(self->conn);
-        close(self->sockfd);
+        self->sockfd = 0;
         Py_END_ALLOW_THREADS
     }
 
@@ -1081,6 +1099,7 @@ PyRabbitMQ_ApplyCallback(PyRabbitMQ_Connection *self,
     PyObject *Message = NULL;
     PyObject *message = NULL;
     PyObject *args = NULL;
+    PyObject *callback_result = NULL;
 
     /* self.callbacks */
     if (!(channel_callbacks = PyDict_GetItem(self->callbacks, channel)))
@@ -1106,11 +1125,15 @@ PyRabbitMQ_ApplyCallback(PyRabbitMQ_Connection *self,
         goto error;
 
     /* callback(message) */
-    Py_INCREF(message);
-    if ((args = PyTuple_New(1)) == NULL) goto finally;
+    if ((args = PyTuple_New(1)) == NULL) {
+        Py_DECREF(message);
+        goto finally;
+    }
     PyTuple_SET_ITEM(args, 0, message);
-    PyObject_CallObject(callback_for_tag, args);
-    Py_DECREF(message);
+
+    callback_result = PyObject_CallObject(callback_for_tag, args);
+    Py_XDECREF(callback_result);
+
     goto finally;
 error:
     retval = -1;
@@ -1276,6 +1299,7 @@ PyRabbitMQ_Connection_queue_bind(PyRabbitMQ_Connection *self,
     PyObject *arguments = NULL;
 
     amqp_table_t bargs = AMQP_EMPTY_TABLE;
+    amqp_pool_t *channel_pool = NULL;
     amqp_rpc_reply_t reply;
 
     if (PyRabbitMQ_Not_Connected(self))
@@ -1288,7 +1312,12 @@ PyRabbitMQ_Connection_queue_bind(PyRabbitMQ_Connection *self,
     if ((exchange = Maybe_Unicode(exchange)) == NULL) goto bail;
     if ((routing_key = Maybe_Unicode(routing_key)) == NULL) goto bail;
 
-    bargs = PyDict_ToAMQTable(self->conn, arguments);
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (channel_pool == NULL) {
+        PyErr_NoMemory();
+        goto bail;
+    }
+    bargs = PyDict_ToAMQTable(self->conn, arguments, channel_pool);
     if (PyErr_Occurred())
         goto bail;
 
@@ -1325,6 +1354,7 @@ PyRabbitMQ_Connection_queue_unbind(PyRabbitMQ_Connection *self,
     PyObject *arguments = NULL;
 
     amqp_table_t uargs = AMQP_EMPTY_TABLE;
+    amqp_pool_t *channel_pool = NULL;
     amqp_rpc_reply_t reply;
 
     if (PyRabbitMQ_Not_Connected(self))
@@ -1336,7 +1366,13 @@ PyRabbitMQ_Connection_queue_unbind(PyRabbitMQ_Connection *self,
     if ((queue = Maybe_Unicode(queue)) == NULL) goto bail;
     if ((exchange = Maybe_Unicode(exchange)) == NULL) goto bail;
     if ((routing_key = Maybe_Unicode(routing_key)) == NULL) goto bail;
-    uargs = PyDict_ToAMQTable(self->conn, arguments);
+
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (channel_pool == NULL) {
+        PyErr_NoMemory();
+        goto bail;
+    }
+    uargs = PyDict_ToAMQTable(self->conn, arguments, channel_pool);
     if (PyErr_Occurred())
         goto bail;
 
@@ -1418,6 +1454,7 @@ PyRabbitMQ_Connection_queue_declare(PyRabbitMQ_Connection *self,
 
     amqp_queue_declare_ok_t *ok;
     amqp_rpc_reply_t reply;
+    amqp_pool_t *channel_pool = NULL;
     amqp_table_t qargs = AMQP_EMPTY_TABLE;
     PyObject *ret = NULL;
 
@@ -1429,7 +1466,12 @@ PyRabbitMQ_Connection_queue_declare(PyRabbitMQ_Connection *self,
             &exclusive, &auto_delete, &arguments))
         goto bail;
     if ((queue = Maybe_Unicode(queue)) == NULL) goto bail;
-    qargs = PyDict_ToAMQTable(self->conn, arguments);
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (channel_pool == NULL) {
+        PyErr_NoMemory();
+        goto bail;
+    }
+    qargs = PyDict_ToAMQTable(self->conn, arguments, channel_pool);
     if (PyErr_Occurred())
         goto bail;
 
@@ -1514,6 +1556,7 @@ PyRabbitMQ_Connection_exchange_declare(PyRabbitMQ_Connection *self,
     unsigned int auto_delete = 0;
 
     amqp_table_t eargs = AMQP_EMPTY_TABLE;
+    amqp_pool_t *channel_pool = NULL;
     amqp_rpc_reply_t reply;
 
     if (PyRabbitMQ_Not_Connected(self))
@@ -1525,7 +1568,12 @@ PyRabbitMQ_Connection_exchange_declare(PyRabbitMQ_Connection *self,
         goto bail;
     if ((exchange = Maybe_Unicode(exchange)) == NULL) goto bail;
     if ((type = Maybe_Unicode(type)) == NULL) goto bail;
-    eargs = PyDict_ToAMQTable(self->conn, arguments);
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (channel_pool == NULL) {
+        PyErr_NoMemory();
+        goto bail;
+    }
+    eargs = PyDict_ToAMQTable(self->conn, arguments, channel_pool);
     if (PyErr_Occurred())
         goto bail;
 
@@ -1603,6 +1651,7 @@ PyRabbitMQ_Connection_basic_publish(PyRabbitMQ_Connection *self,
     int ret = 0;
     amqp_basic_properties_t props;
     amqp_bytes_t bytes;
+    amqp_pool_t *channel_pool = NULL;
     memset(&props, 0, sizeof(props));
 
     if (PyRabbitMQ_Not_Connected(self))
@@ -1616,7 +1665,8 @@ PyRabbitMQ_Connection_basic_publish(PyRabbitMQ_Connection *self,
     if ((routing_key = Maybe_Unicode(routing_key)) == NULL) goto bail;
 
     Py_INCREF(propdict);
-    if (!PyDict_to_basic_properties(propdict, &props, self->conn))
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (!PyDict_to_basic_properties(propdict, &props, self->conn, channel_pool))
         goto bail;
     Py_DECREF(propdict);
 
@@ -1631,6 +1681,7 @@ PyRabbitMQ_Connection_basic_publish(PyRabbitMQ_Connection *self,
                              (amqp_boolean_t)immediate,
                              &props,
                              bytes);
+    amqp_maybe_release_buffers(self->conn);
     Py_END_ALLOW_THREADS;
 
     if (!PyRabbitMQ_HandleError(ret, "basic.publish"))
@@ -1767,6 +1818,7 @@ PyRabbitMQ_Connection_basic_consume(PyRabbitMQ_Connection *self,
 
     amqp_basic_consume_ok_t *ok;
     amqp_rpc_reply_t reply;
+    amqp_pool_t *channel_pool = NULL;
     amqp_table_t cargs = AMQP_EMPTY_TABLE;
 
     if (PyRabbitMQ_Not_Connected(self))
@@ -1779,7 +1831,12 @@ PyRabbitMQ_Connection_basic_consume(PyRabbitMQ_Connection *self,
     if ((queue = Maybe_Unicode(queue)) == NULL) goto bail;
     if ((consumer_tag = Maybe_Unicode(consumer_tag)) == NULL) goto bail;
 
-    cargs = PyDict_ToAMQTable(self->conn, arguments);
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (channel_pool == NULL) {
+        PyErr_NoMemory();
+        goto bail;
+    }
+    cargs = PyDict_ToAMQTable(self->conn, arguments, channel_pool);
     if (PyErr_Occurred()) goto bail;
 
     Py_BEGIN_ALLOW_THREADS;
@@ -1925,7 +1982,7 @@ PyRabbitMQ_Connection_basic_recv(PyRabbitMQ_Connection *self,
 
     if (PyRabbitMQ_recv(self, NULL, self->conn, 0) < 0) {
         if (!PyErr_Occurred())
-            PyErr_SetString(PyRabbitMQExc_ChannelError, "Bad frame read");
+            PyErr_SetString(PyRabbitMQExc_ConnectionError, "Bad frame read");
         goto error;
     }
 
@@ -1994,7 +2051,7 @@ PyRabbitMQ_Connection_basic_get(PyRabbitMQ_Connection *self,
     if (amqp_data_in_buffer(self->conn)) {
         if (PyRabbitMQ_recv(self, p, self->conn, 1) < 0) {
             if (!PyErr_Occurred())
-                PyErr_SetString(PyRabbitMQExc_ChannelError,
+                PyErr_SetString(PyRabbitMQExc_ConnectionError,
                         "Bad frame read");
             Py_XDECREF(p);
             Py_XDECREF(delivery_info);
